# combo - content management system
# Copyright (C) 2014-2017  Entr'ouvert
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from django.conf import settings
from django.db import models
from django.utils.translation import ugettext_lazy as _
from django import template
from django.http import HttpResponse
from django.core.exceptions import PermissionDenied
from django.core.urlresolvers import reverse
from django.forms import models as model_forms, Select
from django.utils.http import quote

from combo.utils import requests
from combo.data.models import CellBase
from combo.data.library import register_cell_class
from combo.utils import NothingInCacheException, get_templated_url

@register_cell_class
class SearchCell(CellBase):
    template_name = 'combo/search-cell.html'

    _search_service = models.CharField(verbose_name=_('Search Service'), max_length=64)

    class Meta:
        verbose_name = _('Search')

    def is_visible(self, user=None):
        if not self.search_service:
            return False
        return super(SearchCell, self).is_visible(user=user)

    def get_default_form_class(self):
        search_services = [(None, _('Not configured'))]
        search_services.append(('_text', _('Page Contents')))
        search_services.extend([(code, service['label'])
                for code, service in settings.COMBO_SEARCH_SERVICES.items()])
        widgets = {'_search_service': Select(choices=search_services)}
        return model_forms.modelform_factory(self.__class__,
                fields=['_search_service'],
                widgets=widgets)

    @property
    def varname(self):
        if self.slug:
            # no hyphen in varname, could be used in context and templates
            return self.slug.replace('-', '_')
        return ''

    @property
    def search_service(self):
        if self._search_service == '_text':
            return {'url': reverse('api-search') + '?q=%(q)s', 'label': _('Page Contents')}
        return settings.COMBO_SEARCH_SERVICES.get(self._search_service) or {}

    def modify_global_context(self, context, request):
        # if self.varname is in the query string (of the page),
        # add it to the global context; so it can be used by others cells
        # for example by a JsonCell with ...[self.varname]... in its URL
        if self.varname and self.varname in request.GET:
            context[self.varname] = request.GET.get(self.varname)

    def get_cell_extra_context(self, context):
        extra_context = super(SearchCell, self).get_cell_extra_context(context)
        # if there is a q_<slug> in query_string, send it to the template (to be
        # used as an initial query) and remove it from query_string
        initial_q = None
        initial_query_string = None
        if context.get('request'):
            request_get = context['request'].GET.copy()
            if self.varname and context.get('request'):
                q_varname = 'q_%s' % self.varname
                if q_varname in request_get:
                    initial_q = request_get[q_varname]
                    del request_get[q_varname]
            initial_query_string = request_get.urlencode()
        extra_context.update({
            'initial_q': initial_q,
            'initial_query_string': initial_query_string
        })
        return extra_context

    @classmethod
    def ajax_results_view(cls, request, cell_pk):
        cell = cls.objects.get(pk=cell_pk)
        if not cell.is_visible(request.user) or not cell.page.is_visible(request.user):
            raise PermissionDenied

        query = request.GET.get('q')
        if query and cell.search_service.get('url'):
            url = cell.search_service.get('url')
            url = get_templated_url(url)
            url = url % {'q': quote(query.encode('utf-8'))}
            if url.startswith('/'):
                url = request.build_absolute_uri(url)
            results = requests.get(url, cache_duration=0).json()
        else:
            results = {'err': 0, 'data': []}

        template_names = ['combo/search-cell-results.html']
        if cell.slug:
            template_names.insert(0, 'combo/cells/%s/search-cell-results.html' % cell.slug)
        tmpl = template.loader.select_template(template_names)
        context= {'cell': cell, 'results': results}
        return HttpResponse(tmpl.render(context, request), content_type='text/html')
